#include "auth.h"

#include "hash/hash.hpp"

#include <print>

#include <leveldb/db.h>
#include <leveldb/write_batch.h>

#include <nlohmann/json.hpp>

using leveldb_ptr = leveldb::DB*;

auto user_db    = leveldb_ptr{nullptr};
auto iter_round = 10000;

extern "C"
{
    auto open_user_db() -> int
    {
        auto opts = leveldb::Options{};

        opts.create_if_missing = true;

        auto status = leveldb::DB::Open(opts, "db/users", &user_db);

        if (!status.ok()) {
            std::println(stderr, "Failed to open user database: {}", status.ToString());
            return 0;
        }
        return 1;
    }

    auto close_user_db() -> void
    {
        delete user_db;
    }

    auto check_user_exists(const char* user_id, int* result) -> int
    {
        auto value  = std::string{};
        auto status = user_db->Get(leveldb::ReadOptions{}, user_id, &value);
        if (status.ok()) {
            *result = 1;
        } else if (status.IsNotFound()) {
            *result = 0;
        } else {
            std::println(stderr, "Failed to check user existence: {}", status.ToString());
            return 0;
        }
        return 1;
    }

    auto set_user_password(const char* user_id, const char* password) -> int
    {
        auto value  = std::string{};
        auto status = user_db->Get(leveldb::ReadOptions{}, user_id, &value);
        if (!status.ok()) {
            std::println(stderr, "Failed to set user password: {}", status.ToString());
            return 0;
        }
        auto json    = nlohmann::json::parse(value);
        json["hash"] = generate_hash(password, iter_round);

        status = user_db->Put(leveldb::WriteOptions{}, user_id, json.dump());
        if (!status.ok()) {
            std::println(stderr, "Failed to set user password: {}", status.ToString());
            return 0;
        }
        return 1;
    }

    auto login(const char* user_id, const char* password, int* result) -> int
    {
        auto value  = std::string{};
        auto status = user_db->Get(leveldb::ReadOptions{}, user_id, &value);
        if (!status.ok()) {
            if (!status.IsNotFound()) std::println(stderr, "Failed to login: {}", status.ToString());
            return 0;
        }
        auto json = nlohmann::json::parse(value);
        *result   = validate_password(password, json["hash"]);
        return 1;
    }

    int registe(const char* user_id, const char* password)
    {
        auto json = nlohmann::json{
            {"hash", generate_hash(password, iter_round)},
            {"permission", 2}
        };
        auto status = user_db->Put(leveldb::WriteOptions{}, user_id, json.dump());
        if (!status.ok()) {
            std::println(stderr, "Failed to register: {}", status.ToString());
            return 0;
        }
        return 1;
    }

    auto delete_user(const char* user_id) -> int
    {
        auto status = user_db->Delete(leveldb::WriteOptions{}, user_id);
        if (!status.ok()) {
            std::println(stderr, "Failed to delete: {}", status.ToString());
            return 0;
        }
        return 1;
    }

    int get_user_permission(const char* user_id, int* result)
    {
        auto value  = std::string{};
        auto status = user_db->Get(leveldb::ReadOptions{}, user_id, &value);
        if (!status.ok()) {
            if (!status.IsNotFound()) std::println(stderr, "Failed to get user permission: {}", status.ToString());
            return 0;
        }
        auto json = nlohmann::json::parse(value);
        json["permission"].get_to(*result);
        return 1;
    }

    int set_user_permission(const char* user_id, int permission)
    {
        auto value  = std::string{};
        auto status = user_db->Get(leveldb::ReadOptions{}, user_id, &value);
        if (!status.ok()) {
            std::println(stderr, "Failed to set user password: {}", status.ToString());
            return 0;
        }
        auto json          = nlohmann::json::parse(value);
        json["permission"] = permission;

        status = user_db->Put(leveldb::WriteOptions{}, user_id, json.dump());
        if (!status.ok()) {
            std::println(stderr, "Failed to set user permission: {}", status.ToString());
            return 0;
        }
        return 1;
    }

    int set_admin_password(const char* password)
    {
        auto json = nlohmann::json{
            {"admin_password_hash", generate_hash(password, iter_round)}
        };
        auto status = user_db->Put(leveldb::WriteOptions{}, "@", json.dump());
        if (!status.ok()) {
            std::println(stderr, "Failed to set admin password: {}", status.ToString());
            return 0;
        }
        return 1;
    }

    int admin_login(const char* password, int* result)
    {
        auto value  = std::string{};
        auto status = user_db->Get(leveldb::ReadOptions{}, "@", &value);
        if (!status.ok()) {
            std::println(stderr, "Failed to login admin: {}", status.ToString());
            return 0;
        }
        auto json = nlohmann::json::parse(value);
        json["admin_password_hash"].get_to(value);
        *result = validate_password(password, value);
        return 1;
    }

    int has_admin_password(int* result)
    {
        auto value  = std::string{};
        auto status = user_db->Get(leveldb::ReadOptions{}, "@", &value);
        if (status.ok()) {
            *result = 1;
        } else if (status.IsNotFound()) {
            *result = 0;
        } else {
            std::println(stderr, "Failed to check admin password existence: {}", status.ToString());
            return 0;
        }
        return 1;
    }
}